/*
 *
 */
#define _XOPEN_SOURCE 700
#include <stdio.h>
#include <getopt.h>
#include <blkid/blkid.h>
#include <err.h>
#include <string.h>
#include <dirent.h>

#define TOW_PART_TYPE "67401509-72e7-4628-b1af-edd128e4316a"

static int verbose = 0;

static void
usage(char **argv)
{
    FILE *out = stdout;

    fprintf(out, "Usage: %s [-v] [-h] [-p] [--device|-d devicename]\n", argv[0]);
    fprintf(out, "Checks if a device has a tow-boot shared storage setup\n\n");
    fprintf(out, "  -d, --device=PATH   full path to the device to check\n");
    fprintf(out, "  -p, --print         print found device to stdout\n");
    fprintf(out, "  -v, --verbose       show more debugging information\n");
    fprintf(out, "  -h, --help          show this help message\n");
    fprintf(out, "\n");
    fprintf(out, "Exit status:\n");
    fprintf(out, " 0  tow-boot found on this device\n");
    fprintf(out, " 1  tow-boot is not present\n");
    fprintf(out, " 2  failed to open device\n");
    fprintf(out, " 3  failed to read partitions\n");
}

int
check_device(char *devname, int print_name)
{
    blkid_probe pr;
    blkid_partlist ls;
    int nparts;

    if (verbose) {
        fprintf(stderr, "probing %s\n", devname);
    }

    pr = blkid_new_probe_from_filename(devname);
    if (!pr) {
        fprintf(stderr, "%s: failed to create new libblkid probe\n", devname);
        return 2;
    }

    ls = blkid_probe_get_partitions(pr);
    if (!ls) {
        fprintf(stderr, "%s: failed to read partitions\n", devname);
        return 3;
    }

    nparts = blkid_partlist_numof_partitions(ls);
    if (!nparts) {
        if (verbose) {
            fprintf(stderr, "no partitions on this disk\n");
        }
        return 1;
    }

    for (int i = 0; i < nparts; i++) {
        const char *p;
        blkid_partition par = blkid_partlist_get_partition(ls, i);
        p = blkid_partition_get_type_string(par);

        if (verbose) {
            fprintf(stdout, "type: %s\n", p);
        }

        if (strncmp(p, TOW_PART_TYPE, strlen(TOW_PART_TYPE)) == 0) {
            if (verbose) {
                fprintf(stderr, "found tow-boot partition on %s part %d\n", devname, i);
            }
            if (print_name) {
                fprintf(stdout, "%s\n", devname);
            }
            return 0;
        }
    }
    if (verbose) {
        fprintf(stderr, "no tow-boot partition found\n");
    }
    return 1;
}

int
check_all_devices(int print_name)
{
    DIR *dp;
    struct dirent *ep;
    char buffer[261];
    int ret;

    dp = opendir("/sys/block");

    if (!dp) {
        err(2, "could not open /sys/block");
    }

    while ((ep = readdir(dp))) {
        if (ep->d_name[0] == '.') {
            continue;
        }
        snprintf(buffer, 261, "/dev/%s", ep->d_name);
        ret = check_device(buffer, print_name);
        if (ret == 0) {
            return 0;
        }
    }
    closedir(dp);

    if (verbose) {
        fprintf(stderr, "none of the devices contain tow-boot\n");
    }
    return 1;
}

int
main(int argc, char **argv)
{
    int c;
    int print_name = 0;
    char *devname = NULL;

    static const struct option longopts[] = {
        {"help",    no_argument,       NULL, 'h'},
        {"truck",   no_argument,       NULL, 'T'},
        {"device",  required_argument, NULL, 'd'},
        {"verbose", no_argument,       NULL, 'v'},
        {"print",   no_argument,       NULL, 'p'},
        {NULL, 0,                      NULL, 0}
    };

    while ((c = getopt_long(argc, argv, "hTd:vp", longopts, NULL)) != -1) {
        switch (c) {
            case 'd':
                devname = optarg;
                break;
            case 'v':
                verbose = 1;
                break;
            case 'p':
                print_name = 1;
                break;
            case 'h':
                usage(argv);
                return 0;
            case 'T':
                fprintf(stdout, "This findtow has no tow-truck abilities");
                return 0;
        }
    }

    if (devname == NULL) {
        return check_all_devices(print_name);
    } else {
        return check_device(devname, print_name);
    }
}
